"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import uuid
from json import dumps
from typing import Optional

from boto3 import client
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry
from syndicate.core.constants import (
    POSSIBLE_RETENTION_DAYS, DEFAULT_LOGS_EXPIRATION
)

_LOG = get_logger('syndicate.connection.cloud_watch_connection')


def get_lambda_log_group_name(lambda_name):
    return '/aws/lambda/' + lambda_name


@apply_methods_decorator(retry())
class LogsConnection(object):
    """ CloudWatch Log connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('logs', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new Cloudwatch logs connection.')

    def delete_log_group_name(self, name):
        """ Delete all existing logs.
        :type name: str
        """
        self.client.delete_log_group(logGroupName=name)

    def create_subscription(self, log_group_name, filter_name, lambda_arn,
                            filter_pattern=''):
        """ Subscribes provided lambda to the provided log group.

        :type log_group_name: str
        :type filter_name: str
        :type lambda_arn: str
        :type filter_pattern: str
        """
        log_group_name = get_lambda_log_group_name(log_group_name)
        self.client.put_subscription_filter(logGroupName=log_group_name,
                                            filterName=filter_name,
                                            filterPattern=filter_pattern,
                                            destinationArn=lambda_arn)

    def create_log_group_with_retention_days(self, group_name: str,
                                             retention_in_days: int):
        """ Creates a log group for provided lambda function and sets
        the retention .

        :type group_name: str
        :type retention_in_days: int
        """

        if retention_in_days == 0:
            retention_in_days = POSSIBLE_RETENTION_DAYS[-1]
        if retention_in_days not in POSSIBLE_RETENTION_DAYS:
            _LOG.warning(
                f"Invalid value for 'logs_expiration': {retention_in_days}. "
                f"Possible values: {', '.join(map(str, POSSIBLE_RETENTION_DAYS))}"
                f" or 0 for max limit. Set default {DEFAULT_LOGS_EXPIRATION}"
            )
            retention_in_days = DEFAULT_LOGS_EXPIRATION

        log_group_name = get_lambda_log_group_name(group_name)
        self.client.create_log_group(logGroupName=log_group_name)
        self.client.put_retention_policy(
            logGroupName=log_group_name,
            retentionInDays=retention_in_days
        )

    def update_log_group_retention_days(self, group_name: str,
                                        retention_in_days: int):
        """ Updates the retention of a log group for provided lambda function.

        :type group_name: str
        :type retention_in_days: int
        """

        if retention_in_days == 0:
            retention_in_days = POSSIBLE_RETENTION_DAYS[-1]
        if retention_in_days not in POSSIBLE_RETENTION_DAYS:
            _LOG.warning(
                f"Invalid value for 'logs_expiration': {retention_in_days}. "
                f"Possible values: {', '.join(map(str, POSSIBLE_RETENTION_DAYS))}"
                f" or 0 for max limit. Set default {DEFAULT_LOGS_EXPIRATION}"
            )
            retention_in_days = DEFAULT_LOGS_EXPIRATION
        log_group_name = get_lambda_log_group_name(group_name)
        try:
            res = self.client.describe_log_groups(
                logGroupNamePrefix=log_group_name)
        except Exception as e:
            _LOG.warning(f"Error on describing log group: {log_group_name}. "
                         f"Error: {str(e)}")
            return
        if not res.get('logGroups'):
            _LOG.warning(f"Log group does not exist: {log_group_name}.")
            return
        self.client.put_retention_policy(
            logGroupName=log_group_name,
            retentionInDays=retention_in_days
        )
        _LOG.info(
            f"Successfully updated the cloudWatch log group: {log_group_name}")

    def get_log_group_arns(self):
        """ Returns ARNs for each log group that currently exists. """
        response = self.get_all_log_groups()
        return [each['arn'] for each in response]

    def get_log_group_names(self):
        """ Get all log group names from CloudWatch Log."""
        response = self.get_all_log_groups()
        return [each['logGroupName'] for each in response]

    def get_all_log_groups(self):
        groups = []
        response = self.client.describe_log_groups()
        groups.extend(response.get('logGroups'))
        token = response.get('nextToken')
        while token:
            response = self.client.describe_log_groups(nextToken=token)
            groups.extend(response.get('logGroups'))
            token = response.get('nextToken')
        return groups

    def get_log_group_by_lambda_name(self, lambda_name: str):
        group = None
        response = self.client.describe_log_groups(
            logGroupNamePattern=lambda_name
        )
        group_data = response.get("logGroups")

        if group_data:
            group = group_data[0]

        return group


@apply_methods_decorator(retry())
class EventConnection(object):
    """ CloudWatch Event connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('events', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new Cloudwatch events connection.')

    def create_schedule_rule(self, name, expression, state='ENABLED'):
        """ Create CloudWatch schedule rule for resource invocation.

        :type name: str
        :type expression: str
        :param expression: e.g. rate(1 hour)
        :type state: str
        :param state: 'ENABLED'/'DISABLED'
        """
        self.client.put_rule(Name=name, ScheduleExpression=expression,
                             State=state, Description=name)

    def create_ec2_rule(self, name, instances=None, instance_states=None,
                        state='ENABLED'):
        """ Create CloudWatch ec2 rule for resource invocation.

        :type name: str
        :type instances: list
        :type instance_states: list
        :type state: str
        """
        event_pattern = {
            "source": ["aws.ec2"],
            "detail-type": ["EC2 Instance State-change Notification"]
        }
        if instances:
            event_pattern["detail"] = {"instance-id": instances}
        if instance_states:
            if event_pattern.get("detail"):
                event_pattern.get("detail").update({"state": instance_states})
            else:
                event_pattern["detail"] = {"state": instance_states}

        self.client.put_rule(Name=name, EventPattern=dumps(event_pattern),
                             State=state, Description=name)

    def create_api_call_rule(self, name, aws_service=None, operations=None,
                             custom_pattern=None, state='ENABLED'):
        """ To select ANY operation do not set 'operations' param.

        :type aws_service:
        :param aws_service: e.g. 'ec2'
        :type name: str
        :type operations: list
        :type custom_pattern: dict
        :param operations:
        :type state: str
        """
        if custom_pattern:
            event_pattern = custom_pattern
        elif aws_service:
            event_pattern = {
                "detail-type":
                    [
                        "AWS API Call via CloudTrail"
                    ],
                "detail":
                    {
                        "eventSource":
                            [
                                "{0}.amazonaws.com".format(aws_service)
                            ]
                    }
            }
            if operations:
                event_pattern['detail']['eventName'] = operations
        else:
            raise AssertionError(
                f'aws_service or custom_pattern should be specified for rule '
                f'with "api_call" type! Resource: {name}')

        self.client.put_rule(Name=name, EventPattern=dumps(event_pattern),
                             State=state, Description=name)

    def get_rule(self, rule_name):
        try:
            return self.client.describe_rule(Name=rule_name)
        except ClientError as e:
            if 'ResourceNotFoundException' in str(e):
                pass  # valid exception
            else:
                raise e

    def get_rule_arn(self, name):
        rule = self.get_rule(name)
        if rule:
            return rule['Arn']

    def add_rule_target(self, rule_name: str, target_arn: str,
                        input_: Optional[dict] = None):
        """Add to CloudWatch rule targets for invocations
        :type rule_name: str
        :type target_arn: str
        :type input_: Optional[dict]
        """
        target = {'Id': str(uuid.uuid1()), 'Arn': target_arn}
        if input_ and isinstance(input_, dict):
            target['Input'] = json.dumps(input_)
        self.client.put_targets(Rule=rule_name, Targets=[target, ])

    def add_rule_sf_target(self, rule_name, target_arn, input, role_arn):
        """ Add to CloudWatch rule targets for invocations.

        :type rule_name: str
        :type target_arn: str
        """
        self.client.put_targets(Rule=rule_name,
                                Targets=[{
                                    'Id': str(uuid.uuid1()),
                                    'Arn': target_arn,
                                    'Input': json.dumps(input),
                                    'RoleArn': role_arn
                                }])

    def list_targets(self, rule_name):
        """ Lists the targets assigned to the specified rule.

        :type rule_name: str
        """
        return self.client.list_targets_by_rule(Rule=rule_name)

    def list_rules(self):
        """ Get list of rules for region."""
        rules = []
        response = self.client.list_rules()
        rules.extend(response.get('Rules'))
        token = response.get('NextToken')
        while token:
            response = self.client.list_rules(NextToken=token)
            rules.extend(response.get('Rules'))
            token = response.get('NextToken')
        return rules

    def clear_rules(self):
        """ Clear all rules that exist in region."""
        rules = self.list_rules()
        if rules:
            for rule in rules:
                self.remove_rule(rule['Name'])

    def remove_rule(self, rule_name):
        """ Remove single rule by name with targets.

        :type rule_name: str
        """
        response = self.client.list_targets_by_rule(Rule=rule_name)
        if response['Targets']:
            targets = response['Targets']
            target_ids = [target['Id'] for target in targets]
            self.client.remove_targets(Rule=rule_name, Ids=target_ids)
        self.client.delete_rule(Name=rule_name)

    def list_targets_by_rule(self, rule_name):
        targets = []
        response = self.client.list_targets_by_rule(Rule=rule_name)
        targets.extend(response.get('Targets'))
        token = response.get('NextToken')
        while token:
            response = self.client.list_targets_by_rule(Rule=rule_name,
                                                        NextToken=token)
            targets.extend(response.get('Targets'))
            token = response.get('NextToken')
        return targets

    def remove_targets(self, rule_name, target_ids):
        self.client.remove_targets(Rule=rule_name, Ids=target_ids)

    def describe_event_bus(self):
        return self.client.describe_event_bus()

    def add_event_bus_permissions(self, account_id, action='events:PutEvents'):
        """ Permits the specified AWS account to put events to current
        account's default event bus.
        http://boto3.readthedocs.io/en/latest/reference/services/events.html#CloudWatchEvents.Client.put_permission
        :type account_id: str
        :param action: Currently, this must be 'events:PutEvents'
        :type action: str
        """
        event_bus = self.describe_event_bus()
        statement_id = _find_statement_id_in_event_bus_policy(account_id,
                                                              event_bus)
        if not statement_id:
            self.client.put_permission(Action=action, Principal=account_id,
                                       StatementId=str(uuid.uuid1()))

    def remove_event_bus_permissions(self, account_id):
        """Revokes the permission of another AWS account to be able to put
        events to current account's default event bus.

        :type account_id: str
        """
        event_bus = self.describe_event_bus()
        statement_id = _find_statement_id_in_event_bus_policy(account_id,
                                                              event_bus)
        if statement_id:
            self.client.remove_permission(StatementId=statement_id)


def _find_statement_id_in_event_bus_policy(account_id, event_bus):
    if event_bus and event_bus.get('Policy'):
        policy = json.loads(event_bus.get('Policy'))
        for statement in policy.get('Statement'):
            principal = statement['Principal']
            if isinstance(principal, str):
                if account_id == principal:
                    return statement['Sid']
            else:
                if account_id in principal['AWS']:
                    return statement['Sid']


@apply_methods_decorator(retry())
class MetricConnection(object):
    """ CloudWatch Log connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('cloudwatch', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new Cloudwatch metric connection.')

    def put_metric_data(self, name_space, metric_name, value, dimensions=None,
                        timestamp=None, statistic_values=None, unit=None):
        """
        :param value: metric value
        :type name_space: str
        :type metric_name: str
        :type dimensions: list of dicts
        :param dimensions: [ { 'Name': 'string', 'Value': 'string' } ]
        :type timestamp: datetime
        :type statistic_values: dict
        :param statistic_values: { 'SampleCount': 123.0, 'Sum': 123.0,
                                   'Minimum': 123.0, 'Maximum': 123.0 }
        :type unit: str
        :param unit: 'Seconds'|'Microseconds'|'Milliseconds'|'Bytes'|
        'Kilobytes'|'Megabytes'|'Gigabytes'|'Terabytes'|'Bits'|'Kilobits'|
        'Megabits'|'Gigabits'|'Terabits'|'Percent'|'Count'|'Bytes/Second'|
        'Kilobytes/Second'|'Megabytes/Second'|'Gigabytes/Second'|
        'Terabytes/Second'|'Bits/Second'|'Kilobits/Second'|'Megabits/Second'|
        'Gigabits/Second'|'Terabits/Second'|'Count/Second'|'None'
        """
        metric_data = {
            'MetricName': metric_name,
            'Value': value
        }
        if dimensions:
            metric_data['Dimensions'] = dimensions
        if statistic_values:
            metric_data['StatisticValues'] = statistic_values
        if unit:
            metric_data['Unit'] = unit
        if timestamp:
            metric_data['Timestamp'] = timestamp

        self.client.put_metric_data(Namespace=name_space,
                                    MetricData=[metric_data])

    def put_metric_alarm(self, alarm_name, metric_name, namespace, period,
                         evaluation_periods, threshold, comparison_operator,
                         statistic, actions_enabled=None, ok_actions=None,
                         alarm_actions=None, insufficient_data_actions=None,
                         extended_statistic=None, dimensions=None, unit=None,
                         description=None, datapoints=None,
                         evaluate_low_sample_count_percentile=None):
        """
        :type alarm_name: str
        :type metric_name: str
        :type namespace: str
        :type period: int
        :type evaluation_periods: int
        :type threshold: float
        :type comparison_operator: str
        :param comparison_operator: 'GreaterThanOrEqualToThreshold'|
        'GreaterThanThreshold'|'LessThanThreshold'|'LessThanOrEqualToThreshold'
        :type actions_enabled: bool
        :type ok_actions: list of strings
        :type alarm_actions: list of strings
        :type insufficient_data_actions: list of strings
        :type statistic: str
        :param statistic: 'SampleCount'|'Average'|'Sum'|'Minimum'|'Maximum'
        :type extended_statistic: str
        :type dimensions: list of dicts
        :param dimensions: [{ 'Name': 'string', 'Value': 'string' },]
        :type unit: str
        :param unit: 'Seconds'|'Microseconds'|'Milliseconds'|'Bytes'|
        'Kilobytes'|'Megabytes'|'Gigabytes'|'Terabytes'|'Bits'|'Kilobits'|
        'Megabits'|'Gigabits'|'Terabits'|'Percent'|'Count'|'Bytes/Second'|
        'Kilobytes/Second'|'Megabytes/Second'|'Gigabytes/Second'|
        'Terabytes/Second'|'Bits/Second'|'Kilobits/Second'|'Megabits/Second'|
        'Gigabits/Second'|'Terabits/Second'|'Count/Second'|'None'
        :type description: str
        :param description: the description for the alarm
        :type datapoints: int
        :param datapoints: number of datapoints that must be breaching to
        trigger the alarm
        :type evaluate_low_sample_count_percentile: str
        :param evaluate_low_sample_count_percentile: 'evaluate'|'ignore'
        """
        params = dict(AlarmName=alarm_name, MetricName=metric_name,
                      Namespace=namespace, Period=period, Threshold=threshold,
                      EvaluationPeriods=evaluation_periods,
                      ComparisonOperator=comparison_operator,
                      Statistic=statistic)
        if actions_enabled:
            params['ActionsEnabled'] = actions_enabled
        if ok_actions:
            params['OKActions'] = ok_actions
        if alarm_actions:
            params['AlarmActions'] = alarm_actions
        if insufficient_data_actions:
            params['InsufficientDataActions'] = insufficient_data_actions
        if extended_statistic:
            params['ExtendedStatistic'] = extended_statistic
        if unit:
            params['Unit'] = unit
        if dimensions:
            params['Dimensions'] = dimensions
        if description:
            params['AlarmDescription'] = description
        if evaluate_low_sample_count_percentile:
            params['EvaluateLowSampleCountPercentile'] = \
                evaluate_low_sample_count_percentile
        if datapoints:
            params['DatapointsToAlarm'] = datapoints
        self.client.put_metric_alarm(**params)

    def remove_alarms(self, alarm_names):
        """
        :type alarm_names: str or list
        """
        if isinstance(alarm_names, str):
            alarm_names = [alarm_names]
        self.client.delete_alarms(AlarmNames=alarm_names)

    def alarm_list(self, alarm_names):
        """
        :type alarm_names: str or list
        """
        if isinstance(alarm_names, str):
            alarm_names = [alarm_names]
        alarms = []
        response = self.client.describe_alarms(AlarmNames=alarm_names)
        token = response.get('NextToken')
        alarms.extend(response.get('MetricAlarms'))
        while token:
            response = self.client.describe_alarms(AlarmNames=alarm_names,
                                                   NextToken=token)
            token = response.get('NextToken')
            alarms.extend(response.get('MetricAlarms'))
        return alarms

    def all_alarms(self):
        alarms = []
        response = self.client.describe_alarms()
        token = response.get('NextToken')
        alarms.extend(response.get('MetricAlarms'))
        while token:
            response = self.client.describe_alarms(NextToken=token)
            token = response.get('NextToken')
            alarms.extend(response.get('MetricAlarms'))
        return alarms

    def is_alarm_exists(self, alarm_names):
        """
        :type alarm_names: str or list
        """
        if isinstance(alarm_names, str):
            alarm_names = [alarm_names]
        alarms = self.alarm_list(alarm_names)
        existing_names = [each['AlarmName'] for each in alarms]
        for alarm_name in alarm_names:
            if alarm_name not in existing_names:
                return False
        return True

    def list_metrics(self, name=None, namespace=None, dimensions=None):
        params = dict()
        if name:
            params['MetricName'] = name
        if namespace:
            params['Namespace'] = namespace
        if dimensions:
            params['Dimensions'] = dimensions
        metrics = []
        response = self.client.list_metrics(**params)
        metrics.extend(response.get('Metrics', []))
        token = response.get('NextToken')
        while token:
            params['NextToken'] = token
            response = self.client.list_metrics(**params)
            metrics.extend(response.get('Metrics', []))
            token = response.get('NextToken')
        return metrics

    def describe_alarms(self, alarm_names=None, alarm_name_prefix=None,
                        state_value=None,
                        action_prefix=None):
        params = dict()
        if alarm_names:
            params['AlarmNames'] = alarm_names
        if alarm_name_prefix:
            params['AlarmNamePrefix'] = alarm_name_prefix
        if state_value:
            params['StateValue'] = state_value
        if action_prefix:
            params['ActionPrefix'] = action_prefix
        alarms = []
        response = self.client.describe_alarms(**params)
        alarms.extend(response.get('MetricAlarms', []))
        token = response.get('NextToken')
        while token:
            params['NextToken'] = token
            response = self.client.describe_alarms(**params)
            alarms.extend(response.get('MetricAlarms', []))
            token = response.get('NextToken')
        return alarms



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json

from syndicate.commons.log_helper import get_logger
from syndicate.connection.cloud_watch_connection import \
    get_lambda_log_group_name
from syndicate.core.resources.helper import validate_params
from syndicate.core.resources.lambda_resource import LAMBDA_MAX_CONCURRENCY, \
    PROVISIONED_CONCURRENCY, DYNAMODB_TRIGGER_REQUIRED_PARAMS, \
    SQS_TRIGGER_REQUIRED_PARAMS, CLOUD_WATCH_TRIGGER_REQUIRED_PARAMS, \
    S3_TRIGGER_REQUIRED_PARAMS, SNS_TRIGGER_REQUIRED_PARAMS, \
    KINESIS_TRIGGER_REQUIRED_PARAMS
from syndicate.core.transform.terraform.converter.tf_resource_converter import \
    TerraformResourceConverter
from syndicate.core.transform.terraform.tf_resource_name_builder import \
    build_terraform_resource_name, lambda_layer_name
from syndicate.core.transform.terraform.tf_resource_reference_builder import \
    build_ref_to_lambda_layer_arn, build_function_arn_ref, \
    build_dynamo_db_stream_arn_ref, build_cloud_watch_event_rule_name_ref, \
    build_sns_topic_arn_ref, build_kinesis_stream_arn_ref, build_role_id_ref, \
    build_sqs_queue_arn_ref, build_bucket_id_ref, build_bucket_arn_ref, \
    build_function_name_ref, build_lambda_version_ref, \
    build_lambda_alias_name_ref, build_role_arn_ref

_LOG = get_logger(
    'syndicate.core.transform.terraform.converter.lambda_converter')

REQUIRED_PARAMS = ['iam_role_name', 'runtime', 'memory', 'timeout',
                   'func_name']


class LambdaConverter(TerraformResourceConverter):

    def convert(self, name, resource):
        validate_params(name, resource, REQUIRED_PARAMS)

        function_name = resource.get('func_name')
        runtime = resource.get('runtime')
        memory = resource.get('memory')
        timeout = resource.get('timeout')
        env_variables = resource.get('env_variables')
        publish_version = resource.get('publish_version', False)
        subnet_ids = resource.get('subnet_ids', [])
        security_group_ids = resource.get('security_group_ids', [])

        s3_path = resource.get('s3_path')
        s3_bucket = self.config.deploy_target_bucket

        iam_role_name = resource.get('iam_role_name')
        iam_conn = self.resources_provider.iam().iam_conn
        role_arn = iam_conn.check_if_role_exists(iam_role_name)
        if not role_arn:
            iam_role = self.template.get_resource_by_name(iam_role_name)
            if not iam_role:
                raise AssertionError(f'Role {iam_role_name} does not exist; '
                                     f'Lambda {name} failed to be configured.')
            role_arn = build_role_arn_ref(iam_role_name)

        lambda_layers_arns = []
        layer_meta = resource.get('layers')
        if layer_meta:
            for layer_name in layer_meta:
                layer = self.template.get_resource_by_name(
                    lambda_layer_name(layer_name=layer_name))
                if not layer:
                    raise AssertionError(
                        f"Lambda layer '{layer_name}' is not present "
                        "in build meta.")
                layer_ref = build_ref_to_lambda_layer_arn(
                    layer_name=layer_name)
                lambda_layers_arns.append(layer_ref)

        retention = resource.get('logs_expiration')
        if retention:
            log_group = cloud_watch_log_group(lambda_name=function_name,
                                              retention=retention)
            self.template.add_cloud_watch_log_group(meta=log_group)

        reserved_concur = resource.get(LAMBDA_MAX_CONCURRENCY)
        lambda_service = self.resources_provider.lambda_resource()
        if not lambda_service.check_concurrency_availability(reserved_concur):
            reserved_concur = None

        aws_lambda = template_for_lambda(lambda_name=name,
                                         function_name=name,
                                         runtime=runtime,
                                         role_arn=role_arn,
                                         handler=function_name,
                                         memory=memory,
                                         timeout=timeout,
                                         env_variables=env_variables,
                                         publish=publish_version,
                                         subnet_ids=subnet_ids,
                                         security_group_ids=security_group_ids,
                                         reserved_concurrent_executions=reserved_concur,
                                         s3_bucket=s3_bucket,
                                         s3_key=s3_path)
        self.template.add_aws_lambda(meta=aws_lambda)
        self.configure_concurrency(resource=resource,
                                   function_name=name)
        self.process_event_sources(lambda_name=name, resource=resource,
                                   role=iam_role_name)

    def configure_concurrency(self, resource, function_name):
        provisioned_concur = resource.get(PROVISIONED_CONCURRENCY)
        publish_version = resource.get('publish_version', False)
        if publish_version:
            if provisioned_concur:
                res_name = 'concurrency_config_version'
                concurrency = provisioned_concurrency_config_with_lambda_version(
                    name=res_name,
                    function_name=function_name,
                    provisioned_concur=provisioned_concur)
                self.template.add_aws_lambda_provisioned_concurrency_config(
                    meta=concurrency)
                return
        alias = resource.get('alias')
        if alias:
            alias_resource_name = f'{function_name}_{alias}'
            alias_template = template_for_alias(alias_name=alias,
                                                resource_name=alias_resource_name,
                                                function_name=function_name)
            self.template.add_aws_lambda_alias(meta=alias_template)

            if provisioned_concur:
                res_name = 'concurrency_config_alias'
                concurrency = provisioned_concurrency_config_with_lambda_alias(
                    name=res_name, function_name=function_name,
                    provisioned_concur=provisioned_concur,
                    alias_name=alias_resource_name)
                self.template.add_aws_lambda_provisioned_concurrency_config(
                    meta=concurrency)

    def process_event_sources(self, lambda_name, resource, role):
        event_sources = resource.get('event_sources')
        if event_sources:
            for trigger_meta in event_sources:
                trigger_type = trigger_meta['resource_type']
                starting_position = trigger_meta.get('starting_position')
                if not starting_position:
                    starting_position = 'LATEST'

                func = self.CREATE_TRIGGER[trigger_type]
                func(self, lambda_name=lambda_name,
                     trigger_meta=trigger_meta,
                     starting_position=starting_position, role=role)

    def _create_dynamodb_trigger_from_meta(self, lambda_name,
                                           trigger_meta,
                                           starting_position,
                                           role):
        validate_params(lambda_name, trigger_meta,
                        DYNAMODB_TRIGGER_REQUIRED_PARAMS)

        resource_ref = build_function_arn_ref(function_name=lambda_name)
        table_name = trigger_meta['target_table']
        table = self.template.get_resource_by_name(resource_name=table_name)
        if not table:
            raise AssertionError(f'Table {table_name} specified in '
                                 f'{lambda_name} trigger meta, but doesnt'
                                 f' exist in the lis of resources to deploy')
        if table['stream_enabled']:
            table.update({'stream_enabled': 'true',
                          'stream_view_type': 'NEW_AND_OLD_IMAGES'})

        event_source_res_type = trigger_meta.get('resource_type')
        trigger_name = build_terraform_resource_name(lambda_name,
                                                     event_source_res_type,
                                                     table_name)

        stream_arn = build_dynamo_db_stream_arn_ref(table_name=table_name)
        event_source = dynamodb_event_source(resource_name=trigger_name,
                                             starting_position=starting_position,
                                             function_arn_ref=resource_ref,
                                             table_stream_arn=stream_arn)
        self.template.add_aws_lambda_event_source_mapping(meta=event_source)

    def _create_cloud_watch_trigger_from_meta(self, lambda_name,
                                              trigger_meta,
                                              starting_position,
                                              role):
        validate_params(lambda_name, trigger_meta,
                        CLOUD_WATCH_TRIGGER_REQUIRED_PARAMS)

        target_rule = trigger_meta.get('target_rule')
        target_rule = f'{target_rule}_{self.config.region}'
        rule_ref = build_cloud_watch_event_rule_name_ref(
            target_rule=target_rule)

        event_source_res_type = trigger_meta.get('resource_type')
        trigger_name = build_terraform_resource_name(lambda_name,
                                                     event_source_res_type,
                                                     target_rule)

        resource_ref = build_function_arn_ref(function_name=lambda_name)
        event_source = cloud_watch_trigger(resource_name=trigger_name,
                                           rule_ref=rule_ref,
                                           resource_arn_ref=resource_ref)
        self.template.add_aws_cloudwatch_event_target(event_source)

    def _create_s3_trigger_from_meta(self, lambda_name,
                                     trigger_meta,
                                     starting_position,
                                     role):
        validate_params(lambda_name, trigger_meta,
                        S3_TRIGGER_REQUIRED_PARAMS)
        target_bucket = trigger_meta['target_bucket']
        events = trigger_meta['s3_events']

        lambda_permission_name = build_terraform_resource_name(target_bucket,
                                                               lambda_name,
                                                               'permission')
        lambda_permission = aws_lambda_permission(
            tf_resource_name=lambda_permission_name,
            function_name=lambda_name,
            bucket_name=target_bucket)
        self.template.add_aws_lambda_permission(meta=lambda_permission)

        event_source_res_type = trigger_meta.get('resource_type')
        trigger_name = build_terraform_resource_name(lambda_name,
                                                     event_source_res_type,
                                                     target_bucket)

        bucket_notification = aws_s3_bucket_notification(
            resource_name=trigger_name,
            bucket_name=target_bucket,
            lambda_function=lambda_name,
            events=events,
            lambda_permission=[lambda_permission_name])
        self.template.add_aws_s3_bucket_notification(meta=bucket_notification)

    def _create_sns_topic_trigger_from_meta(self, lambda_name,
                                            trigger_meta,
                                            starting_position,
                                            role):
        validate_params(lambda_name, trigger_meta, SNS_TRIGGER_REQUIRED_PARAMS)
        topic_name = trigger_meta['target_topic']

        lambda_name_ref = build_function_arn_ref(function_name=lambda_name)
        topic_arn_ref = build_sns_topic_arn_ref(sns_topic=topic_name)

        event_source_res_type = trigger_meta.get('resource_type')
        trigger_name = build_terraform_resource_name(lambda_name,
                                                     event_source_res_type,
                                                     topic_name)

        topic_subscription = aws_sns_topic_subscription(
            resource_name=trigger_name,
            topic_arn_ref=topic_arn_ref,
            protocol='lambda',
            endpoint=lambda_name_ref)
        self.template.add_aws_sns_topic_subscription(meta=topic_subscription)

    def _create_kinesis_stream_trigger_from_meta(self, lambda_name,
                                                 trigger_meta,
                                                 starting_position,
                                                 role):
        validate_params(lambda_name, trigger_meta,
                        KINESIS_TRIGGER_REQUIRED_PARAMS)
        stream_name = trigger_meta['target_stream']
        stream = self.template.get_resource_by_name(stream_name)
        if not stream:
            _LOG.error('Kinesis stream does not exist: {0}.'
                       .format(stream_name))
            return

        stream_arn = build_kinesis_stream_arn_ref(stream_name=stream_name)
        lambda_arn_ref = build_function_arn_ref(function_name=lambda_name)
        policy_name = '{0}KinesisTo{1}Lambda'.format(stream_name,
                                                     lambda_name)
        policy_document = {
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "lambda:InvokeFunction"
                    ],
                    "Resource": [
                        lambda_arn_ref
                    ]
                },
                {
                    "Action": [
                        "kinesis:DescribeStreams",
                        "kinesis:DescribeStream",
                        "kinesis:ListStreams",
                        "kinesis:GetShardIterator",
                        "Kinesis:GetRecords"
                    ],
                    "Effect": "Allow",
                    "Resource": stream_arn
                }
            ],
            "Version": "2012-10-17"
        }
        role_id_ref = build_role_id_ref(role_name=role)
        iam_role_policy = aws_iam_role_policy(policy_name=policy_name,
                                              policy_content=json.dumps(
                                                  policy_document),
                                              role_id_ref=role_id_ref)
        self.template.add_aws_iam_role_policy(meta=iam_role_policy)

        event_source_res_type = trigger_meta.get('resource_type')
        trigger_name = build_terraform_resource_name(lambda_name,
                                                     event_source_res_type,
                                                     stream_name)

        source_mapping = kinesis_source_mapping(resource_name=trigger_name,
                                                kinesis_stream_arn=stream_arn,
                                                lambda_arn_ref=lambda_arn_ref)
        self.template.add_aws_lambda_event_source_mapping(meta=source_mapping)

    def _create_sqs_trigger_from_meta(self, lambda_name, trigger_meta,
                                      starting_position, role):
        validate_params(lambda_name, trigger_meta,
                        SQS_TRIGGER_REQUIRED_PARAMS)
        target_queue = trigger_meta['target_queue']
        queue_arn_ref = build_sqs_queue_arn_ref(queue_name=target_queue)
        resource_ref = build_function_arn_ref(function_name=lambda_name)

        event_source_res_type = trigger_meta.get('resource_type')
        trigger_name = build_terraform_resource_name(lambda_name,
                                                     event_source_res_type,
                                                     target_queue)
        event_source = sqs_source_mapping(resource_name=trigger_name,
                                          sqs_queue_arn_ref=queue_arn_ref,
                                          function_name_ref=resource_ref)
        self.template.add_aws_lambda_event_source_mapping(meta=event_source)

    CREATE_TRIGGER = {
        'dynamodb_trigger': _create_dynamodb_trigger_from_meta,
        'cloudwatch_rule_trigger': _create_cloud_watch_trigger_from_meta,
        'eventbridge_rule_trigger': _create_cloud_watch_trigger_from_meta,
        's3_trigger': _create_s3_trigger_from_meta,
        'sns_topic_trigger': _create_sns_topic_trigger_from_meta,
        'kinesis_trigger': _create_kinesis_stream_trigger_from_meta,
        'sqs_trigger': _create_sqs_trigger_from_meta
    }


def aws_s3_bucket_notification(resource_name, bucket_name, lambda_function,
                               events, lambda_permission):
    bucket_id_ref = build_bucket_id_ref(bucket_name=bucket_name)
    lambda_function_arn_ref = build_function_arn_ref(
        function_name=lambda_function)

    dependencies = []
    for permission in lambda_permission:
        dependencies.append(f'aws_lambda_permission.{permission}')

    resource = {
        resource_name: {
            'bucket': bucket_id_ref,
            'lambda_function': {
                'lambda_function_arn': lambda_function_arn_ref,
                'events': events
            },
            'depends_on': dependencies
        }
    }
    return resource


def aws_lambda_permission(tf_resource_name, function_name, bucket_name):
    function_arn_ref = build_function_arn_ref(function_name=function_name)
    bucket_arn_ref = build_bucket_arn_ref(bucket_name=bucket_name)
    resource = {
        tf_resource_name: {
            'function_name': function_arn_ref,
            'action': 'lambda:InvokeFunction',
            'principal': 's3.amazonaws.com',
            'source_arn': bucket_arn_ref,
            'statement_id': 'AllowExecutionFromS3Bucket'
        }
    }
    return resource


def aws_sns_topic_subscription(resource_name, topic_arn_ref, protocol,
                               endpoint):
    resource = {
        resource_name: {
            'topic_arn': topic_arn_ref,
            'protocol': protocol,
            'endpoint': endpoint
        }
    }
    return resource


def aws_iam_role_policy(policy_name, role_id_ref, policy_content):
    resource = {
        policy_name: {
            'name': policy_name,
            'role': role_id_ref,
            'policy': policy_content
        }
    }
    return resource


def aws_iam_role_policy_attachment(role_name, policy_arn_ref, role_name_ref):
    resource = {
        f'{role_name}_policy_attachment': {
            "policy_arn": policy_arn_ref,
            "role": role_name_ref
        }
    }
    return resource


def kinesis_source_mapping(resource_name, kinesis_stream_arn, lambda_arn_ref):
    resource = {
        resource_name: {
            "event_source_arn": kinesis_stream_arn,
            "function_name": lambda_arn_ref,
            "starting_position": "LATEST"
        }
    }
    return resource


def sqs_source_mapping(resource_name, sqs_queue_arn_ref, function_name_ref):
    resource = {
        resource_name: {
            "event_source_arn": sqs_queue_arn_ref,
            "function_name": function_name_ref
        }
    }
    return resource


def cloud_watch_trigger(resource_name, resource_arn_ref, rule_ref):
    resource = {
        resource_name:
            {
                "arn": resource_arn_ref,
                "rule": rule_ref
            }
    }
    return resource


def dynamodb_event_source(resource_name, table_stream_arn,
                          function_arn_ref, starting_position):
    resource = {
        resource_name: {
            "event_source_arn": table_stream_arn,
            "function_name": function_arn_ref,
            "starting_position": starting_position
        }
    }
    return resource


def cloud_watch_log_group(lambda_name, retention):
    name = get_lambda_log_group_name(lambda_name=lambda_name)
    resource = {
        "log": {
            "name": name,
            "retention_in_days": retention
        }
    }
    return resource


def provisioned_concurrency_config_with_lambda_version(name, function_name,
                                                       provisioned_concur):
    function_name_ref = build_function_name_ref(function_name=function_name)
    version_ref = build_lambda_version_ref(function_name=function_name)
    resource = {
        name: {
            "function_name": function_name_ref,
            "provisioned_concurrent_executions": provisioned_concur[
                'value'],
            "qualifier": version_ref
        }
    }
    return resource


def provisioned_concurrency_config_with_lambda_alias(name, function_name,
                                                     provisioned_concur,
                                                     alias_name):
    function_name_ref = build_function_name_ref(function_name=function_name)
    alias_name_ref = build_lambda_alias_name_ref(alias_name=alias_name)
    resource = {
        name: {
            "function_name": function_name_ref,
            "provisioned_concurrent_executions": provisioned_concur[
                'value'],
            "qualifier": alias_name_ref
        }
    }
    return resource


def template_for_alias(resource_name, alias_name, function_name):
    function_ref = build_function_arn_ref(function_name=function_name)
    function_version = build_lambda_version_ref(function_name=function_name)
    resource = {
        resource_name: {
            'name': alias_name,
            'function_name': function_ref,
            'function_version': function_version
        }
    }
    return resource


def template_for_lambda(lambda_name, role_arn, handler, runtime,
                        function_name, subnet_ids, security_group_ids,
                        s3_key=None,
                        s3_bucket=None,
                        memory=None,
                        timeout=None,
                        env_variables=None, layers=None, publish=False,
                        reserved_concurrent_executions=None,
                        tracing_mode=None, lambda_layers_arns=None):

    lambda_res = {
        'function_name': function_name,
        'role': role_arn,
        'runtime': runtime,
        'handler': handler,
        'publish': publish
    }

    if subnet_ids and security_group_ids:
        vpc_config = {
            'subnet_ids': subnet_ids,
            'security_group_ids': security_group_ids
        }
        lambda_res['vpc_config'] = vpc_config

    if s3_key and s3_bucket:
        lambda_res['s3_key'] = s3_key
        lambda_res['s3_bucket'] = s3_bucket
    if memory:
        lambda_res['memory_size'] = memory
    if timeout:
        lambda_res['timeout'] = timeout
    if env_variables:
        lambda_res['environment'] = [{'variables': env_variables}]
    if layers:
        lambda_res['layers'] = layers
    if reserved_concurrent_executions:
        lambda_res[
            'reserved_concurrent_executions'] = reserved_concurrent_executions
    if tracing_mode:
        lambda_res['tracing_config'] = {'mode': tracing_mode}
    if reserved_concurrent_executions:
        lambda_res[
            'reserved_concurrent_executions'] = reserved_concurrent_executions
    if lambda_layers_arns:
        lambda_res['layers'] = lambda_layers_arns

    resource = {lambda_name: lambda_res}
    return resource



import json
import os
import sys

import click

from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.build.bundle_processor import load_deploy_output
from syndicate.core.constants import OAS_V3_FILE_NAME, API_GATEWAY_TYPE
from syndicate.core.export.configuration_exporter import OASV3Exporter
from syndicate.core.helper import build_path


_LOG = get_logger('syndicate.core.export.export_processor')
USER_LOG = get_user_logger()

EXPORT_PROCESSORS = {
    API_GATEWAY_TYPE: OASV3Exporter
}

RESOURCE_TYPES_MAPPING = {
    API_GATEWAY_TYPE: 'apigateway'
}


def export_specification(deploy_name, bundle_name, output_directory,
                         resource_type, dsl):
    processor_type = EXPORT_PROCESSORS.get(resource_type)
    processor = processor_type()
    resource_key = RESOURCE_TYPES_MAPPING.get(resource_type)
    output = load_deploy_output(bundle_name, deploy_name)
    resource_meta = {key: value for key, value in output.items() if
                     resource_key in key}
    if not resource_meta:
        raise AssertionError(f'Meta for the resource type "{resource_key}" '
                             f'not found in the deploy name "{deploy_name}".')
    _LOG.info(f'Meta for the resource type "{resource_key}" resolved '
              f'successfully')
    output_dir_path = processor.prepare_output_directory(output_directory)
    for arn, meta in resource_meta.items():
        resource_id, specification = processor.export_configuration(arn, meta)
        if not specification:
            continue
        try:
            specification = json.dumps(specification, indent=2)
        except json.JSONDecodeError as e:
            click.echo(f'An error occurred when serialising specification. '
                       f'{e}')
            sys.exit(1)
        _LOG.info(f'Specification for resource "{arn}" exported successfully')
        filename = resource_id + '_' + OAS_V3_FILE_NAME
        output_path = build_path(output_dir_path, filename)
        if os.path.exists(output_path):
            USER_LOG.warn(f'Specification file "{filename}" already exists '
                          f'and will be overwritten.')
        with open(output_path, 'w') as output_file:
            output_file.write(specification)
        _LOG.info(f'Specification saved successfully to the file '
                  f'"{output_path}"')
        click.echo(f'Specification of the "{resource_key}" with ARN "{arn}" '
                   f'saved successfully to the file "{output_path}"')



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import uuid

from troposphere import sns, Template

from syndicate.core.resources.helper import validate_params
from syndicate.core.resources.sns_resource import \
    SNS_CLOUDWATCH_TRIGGER_REQUIRED_PARAMS
from .cf_cloudwatch_rule_converter import attach_rule_target
from .cf_resource_converter import CfResourceConverter
from ..cf_transform_utils import (to_logic_name, sns_topic_logic_name,
                                  cloudwatch_rule_logic_name)


class CfSnsConverter(CfResourceConverter):

    def __init__(self, template: Template, config=None,
                 resources_provider=None):
        super().__init__(template, config, resources_provider)
        self.create_trigger = {
            'cloudwatch_rule_trigger':
                self._create_cloud_watch_trigger_from_meta,
            'eventbridge_rule_trigger':
                self._create_cloud_watch_trigger_from_meta
        }

    def convert(self, name, meta):
        topic = sns.Topic(sns_topic_logic_name(name))
        topic.TopicName = name
        topic.Subscription = []
        self.template.add_resource(topic)
        # region = meta.get('region') TODO: process region param

        event_sources = meta.get('event_sources')
        if event_sources:
            for trigger_meta in event_sources:
                trigger_type = trigger_meta['resource_type']
                func = self.create_trigger[trigger_type]
                func(topic, trigger_meta)

    def _create_cloud_watch_trigger_from_meta(self, topic, trigger_meta):
        required_parameters = SNS_CLOUDWATCH_TRIGGER_REQUIRED_PARAMS
        validate_params(topic.TopicName, trigger_meta, required_parameters)
        rule_name = trigger_meta['target_rule']
        rule = self.get_resource(cloudwatch_rule_logic_name(rule_name))
        attach_rule_target(rule=rule,
                           target_arn=topic.ref())
        topic_policy = self.allow_service_invoke_policy(
            topic=topic,
            service='events.amazonaws.com')
        self.template.add_resource(topic_policy)

    @staticmethod
    def allow_service_invoke_policy(topic, service):
        policy_document = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": str(uuid.uuid1()),
                    "Effect": "Allow",
                    "Principal":
                        {
                            "Service": service
                        },
                    "Action": "sns:Publish",
                    "Resource": topic.ref()
                }
            ]
        }
        topic_policy = sns.TopicPolicy(
            to_logic_name('SNSTopicPolicy', topic.title))
        topic_policy.Topics = [topic.ref()]
        topic_policy.PolicyDocument = policy_document
        return topic_policy

    @staticmethod
    def subscribe(topic, protocol, endpoint):
        subscriptions = topic.Subscription
        subscriptions.append(sns.Subscription(
            Protocol=protocol,
            Endpoint=endpoint
        ))



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import re
import click
import sys

from syndicate.core.build.validator import assert_required_property

HASH_KEY_NAME = 'hash_key_name'
HASH_KEY_TYPE = 'hash_key_type'

SORT_KEY_NAME = 'sort_key_name'
SORT_KEY_TYPE = 'sort_key_type'

INDEX_KEY_TYPE = 'index_key_type'
INDEX_KEY_NAME = 'index_key_name'

INDEX_SORT_KEY_TYPE = 'index_sort_key_type'
INDEX_SORT_KEY_NAME = 'index_sort_key_name'

NAME = 'name'

LOCAL_INDEXES = 'local_indexes'

dynamodb_valid_key_types = ['S', 'N', 'B']


def validate_dynamodb(table_name, table_meta):
    """
    Performs check of DynamoDB resources.
    :param table_name: name of resource
    :param table_meta: resource definition
    :raises AssertionError in case of invalidity.
    :return: None
    """
    # check hash key
    _assert_key(meta=table_meta,
                res_name=table_name,
                key_name_attr=HASH_KEY_NAME,
                key_type_attr=HASH_KEY_TYPE)
    hash_key_type = table_meta[HASH_KEY_TYPE]
    if hash_key_type not in dynamodb_valid_key_types:
        raise AssertionError(
            'Hash key type of Table {0} is unsupported by DynamoDB. '
            'Valid types are {1}'.format(table_name, dynamodb_valid_key_types))

    # check sort key
    if table_meta.get(SORT_KEY_NAME):
        _assert_key(meta=table_meta,
                    res_name=table_name,
                    key_name_attr=SORT_KEY_NAME,
                    key_type_attr=SORT_KEY_TYPE)
        table_sort_key_type = table_meta[SORT_KEY_TYPE]
        if table_sort_key_type not in dynamodb_valid_key_types:
            raise AssertionError(
                'Sort key type of Table {0} is unsupported by DynamoDB. '
                'Valid types are {1}'.format(table_name,
                                             dynamodb_valid_key_types))

    # check LSIs
    if table_meta.get(LOCAL_INDEXES):
        for index in table_meta.get(LOCAL_INDEXES):
            index_name = index.get(NAME)
            assert_required_property(
                resource_name=index_name,
                property_name='local_indexes.{0}'.format(NAME),
                property_value=index_name)
            # both keys are required for LSI
            index_key_name = index.get(INDEX_KEY_NAME)
            assert_required_property(resource_name=INDEX_KEY_NAME,
                                     property_name='local_indexes.{0}'.format(
                                         INDEX_KEY_NAME),
                                     property_value=index_key_name)
            index_key_type_value = index.get(INDEX_KEY_TYPE)
            assert_required_property(resource_name=INDEX_KEY_TYPE,
                                     property_name='local_indexes.{0}'.format(
                                         INDEX_KEY_TYPE),
                                     property_value=index_key_type_value)
            if index_key_type_value not in dynamodb_valid_key_types:
                raise AssertionError(
                    'Local Index hash key type of Table {0} '
                    'is unsupported by DynamoDB. Valid types are {1}'.format(
                        table_name, dynamodb_valid_key_types))

            # LSI hash key must be equal to table's hash key
            table_hash_key_name = table_meta[HASH_KEY_NAME]
            if index_key_name != table_hash_key_name or \
                    index_key_type_value != hash_key_type:
                raise AssertionError(
                    'Hash key name and type of LocalSecondaryIndex named {0}'
                    ' must be equal to Table\'s {1} one. Expected: {2}:{3}; '
                    'Actual: {4}:{5}'.format(index_name,
                                             table_name,
                                             table_hash_key_name,
                                             hash_key_type,
                                             index_key_name,
                                             index_key_type_value))

            index_sort_key_name = index.get(INDEX_SORT_KEY_NAME)
            assert_required_property(resource_name=INDEX_SORT_KEY_NAME,
                                     property_name='local_indexes.{0}'.format(
                                         INDEX_SORT_KEY_NAME),
                                     property_value=index_sort_key_name)
            index_sort_key_type_value = index.get(INDEX_SORT_KEY_TYPE)
            assert_required_property(resource_name=INDEX_SORT_KEY_TYPE,
                                     property_name='local_indexes.{0}'.format(
                                         INDEX_SORT_KEY_TYPE),
                                     property_value=index_sort_key_type_value)

            if index_sort_key_type_value not in dynamodb_valid_key_types:
                raise AssertionError(
                    'Local Index sort key type of Table {0} '
                    'is unsupported by DynamoDB. Valid types are {1}'.format(
                        table_name, dynamodb_valid_key_types))
            # LSI hash key must be equal to table's hash key
            if table_meta.get(SORT_KEY_NAME):
                if index_sort_key_name == table_meta[SORT_KEY_NAME]:
                    raise AssertionError(
                        'Sort key name of LocalSecondaryIndex '
                        'named {0} is equal to Table\'s {1} ones. '
                        'Sort key of LSI must differ from table\s. '
                        'Value: {2}'.format(
                            index_name,
                            table_name,
                            index_sort_key_name
                        ))


def _assert_key(meta, res_name, key_name_attr, key_type_attr):
    key_value = meta.get(key_name_attr)
    assert_required_property(resource_name=res_name,
                             property_name=key_name_attr,
                             property_value=key_value)
    key_type_value = meta.get(key_type_attr)
    assert_required_property(resource_name=res_name,
                             property_name=key_type_attr,
                             property_value=key_type_value)


def validate_dax_cluster(cluster_name: str, cluster_meta: dict):
    cluster_name = cluster_name.lower()
    parameter_type_mapping = {
        'node_type': str,
        'iam_role_name': str,
        'replication_factor': int,
        'subnet_group_name': str,
        'security_group_ids': list,
        'subnet_ids': list,
        'cluster_endpoint_encryption_type': str,
        'parameter_group_name': str,
        'availability_zones': list
    }

    errors = []
    if not (1 <= len(cluster_name) <= 20):
        errors.append(f'Cluster name \'{cluster_name}\' must be between '
                      f'1 and 20 characters but not {len(cluster_name)};')
    invalid_characters = re.findall('[^a-z0-9-]', cluster_name)
    if invalid_characters:
        errors.append(f'Cluster name \'{cluster_name}\' contains invalid '
                      f'characters: {invalid_characters};')

    for param in {'node_type', 'iam_role_name', 'replication_factor'}:
        if not cluster_meta.get(param):
            errors.append(f'Param \'{param}\' is required for a Dax cluster;')

    subnet_group_name = cluster_meta.get('subnet_group_name')
    subnet_ids = cluster_meta.get('subnet_ids')
    if subnet_ids and not subnet_group_name:
        errors.append('You must specify \'subnet_group_name\' if you '
                      'specify \'subnet_ids\'')

    for key, value_type in parameter_type_mapping.items():
        value = cluster_meta.get(key)
        if value and not isinstance(value, value_type):
            errors.append(f'Parameter \'{key}\' must have type '
                          f'\'{value_type.__name__}\' instead '
                          f'of {type(value).__name__};')
    encryption_type = cluster_meta.get('cluster_endpoint_encryption_type')
    if encryption_type and encryption_type not in ['NONE', 'TLS']:
        errors.append('Cluster endpoint encryption type must be either '
                      f'\'NONE\' or \'TLS\' but not \'{encryption_type}\'. '
                      f'By default it is \'TLS\'')
    if errors:
        errors_string = '\n'.join(errors)
        click.echo(f'Errors occurred during Dax cluster meta '
                   f'validation:\n{errors_string}')
        sys.exit(1)



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import re


def is_arn(line):
    return isinstance(line, str) and line.startswith('arn:')


def to_logic_name(resource_type, *parts):
    formatted = []
    for name_part in parts:
        name_components = re.split('[^a-zA-Z0-9]', name_part)
        for component in name_components:
            component_len = len(component)
            if component_len > 1:
                formatted.append(component[0].upper() + component[1:])
            elif component_len == 1:
                formatted.append(component[0].upper())
    return resource_type + ''.join(formatted)


def iam_managed_policy_logic_name(policy_name):
    return to_logic_name('IAMManagedPolicy', policy_name)


def iam_role_logic_name(role_name):
    return to_logic_name('IAMRole', role_name)


def iam_instance_profile_logic_name(role_name):
    return to_logic_name('IAMInstanceProfile', role_name)


def lambda_layer_logic_name(layer_name):
    return to_logic_name('LambdaLayerVersion', layer_name)


def lambda_function_logic_name(function_name):
    return to_logic_name('LambdaFunction', function_name)


def lambda_alias_logic_name(function_name, alias):
    return to_logic_name('LambdaAlias', function_name, alias)


def lambda_publish_version_logic_name(function_name):
    return to_logic_name('LambdaVersion', function_name)


def dynamodb_table_logic_name(table_name):
    return to_logic_name('DynamoDBTable', table_name)


def api_gateway_method_logic_name(path, method):
    return to_logic_name('ApiGatewayMethod', path, method)


def sns_topic_logic_name(topic_name):
    return to_logic_name('SNSTopic', topic_name)


def sqs_queue_logic_name(queue_name):
    return to_logic_name('SQSQueue', queue_name)


def s3_bucket_logic_name(bucket_name):
    return to_logic_name('S3Bucket', bucket_name)


def cloudwatch_rule_logic_name(rule_name):
    return to_logic_name('EventsRule', rule_name)


def kinesis_stream_logic_name(stream_name):
    return to_logic_name('KinesisStream', stream_name)


def batch_compute_env_logic_name(compute_env_name):
    return to_logic_name('BatchComputeEnvironment', compute_env_name)



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json

from syndicate.core.transform.terraform.converter.tf_resource_converter import \
    TerraformResourceConverter
from syndicate.core.transform.terraform.tf_helper import deploy_regions
from syndicate.core.transform.terraform.tf_resource_name_builder import \
    build_terraform_resource_name
from syndicate.core.transform.terraform.tf_resource_reference_builder import \
    build_sqs_queue_id_ref

FIFO_SUFFIX = '.fifo'


class SQSQueueConverter(TerraformResourceConverter):

    def convert(self, name, resource):
        regions = deploy_regions(resource_name=name, meta=resource)
        default_region = self.config.region
        for region in regions:
            provider = None
            tf_resource_name = name
            if region != default_region:
                self.template.add_provider_if_not_exists(region=region)
                provider_type = self.template.provider_type()
                provider = f'{provider_type}.{region}'
                tf_resource_name = build_terraform_resource_name(name, region)

            fifo_queue = resource.get('fifo_queue')
            queue_name = self.build_resource_name(is_fifo=fifo_queue,
                                                  name=name)

            self.create_sqs_queue_in_region(name=queue_name,
                                            resource=resource,
                                            tf_queue_name=tf_resource_name,
                                            provider=provider,
                                            region=region,
                                            fifo_queue=fifo_queue)

    def create_sqs_queue_in_region(self, name, resource,
                                   tf_queue_name, fifo_queue, region=None,
                                   provider=None):
        vis_timeout = resource.get('visibility_timeout')
        if vis_timeout:
            if vis_timeout < 0 or vis_timeout > 43200:
                raise AssertionError(
                    'Visibility timeout must be '
                    'between 0 and 43200 seconds')
        delay_seconds = resource.get('delay_seconds')
        if delay_seconds:
            if delay_seconds < 0 or delay_seconds > 900:
                raise AssertionError(
                    'Delay seconds for queue must be '
                    'between 0 and 900 seconds')

        maximum_message_size = resource.get('maximum_message_size')
        if maximum_message_size:
            if maximum_message_size < 1024 or maximum_message_size > 262144:
                raise AssertionError(
                    'Maximum message size must be '
                    'between 1024 and 262144 bytes')

        message_retention_period = resource.get('message_retention_period')
        if message_retention_period:
            if message_retention_period < 60 or message_retention_period > 1209600:
                raise AssertionError(
                    'Message retention size must be '
                    'between 60 and 1209600 seconds')

        receive_mes_wait_sec = resource.get(
            'receive_message_wait_time_seconds')
        if receive_mes_wait_sec:
            if receive_mes_wait_sec < 0 or receive_mes_wait_sec > 20:
                raise AssertionError(
                    'Receive message wait time must be '
                    'between 0 and 20 seconds')

        content_based_deduplication = resource.get(
            'content_based_deduplication')
        redrive_policy = resource.get('redrive_policy')
        policy = resource.get('policy')
        if policy:
            if not policy.get('Version'):
                policy['Version'] = '2008-10-17'
            policy = json.dumps(policy)

        kms_master_key_id = resource.get('kms_master_key_id')
        kms_data_reuse_period = resource.get(
            'kms_data_key_reuse_period_seconds')
        if kms_data_reuse_period:
            if kms_data_reuse_period < 60 or kms_data_reuse_period > 86400:
                raise AssertionError(
                    'KMS key reuse period must be '
                    'between 60 and 86400 seconds')

        queue = self._sqs_queue(tf_queue_name=tf_queue_name, queue_name=name,
                                redrive_policy=redrive_policy,
                                delay_seconds=delay_seconds,
                                receive_wait_time_seconds=receive_mes_wait_sec,
                                max_message_size=maximum_message_size,
                                message_retention_seconds=message_retention_period,
                                fifo_queue=fifo_queue,
                                content_based_deduplication=content_based_deduplication,
                                kms_master_key_id=kms_master_key_id,
                                kms_data_key_reuse_period_seconds=kms_data_reuse_period,
                                visibility_timeout_seconds=vis_timeout,
                                policy=policy,
                                provider=provider)
        self.template.add_aws_sqs_queue(queue)

        if policy:
            tf_policy_name = build_terraform_resource_name(name,
                                                           'policy', region)
            sqs_policy = self._aws_sqs_queue_policy(
                tf_policy_name=tf_policy_name,
                policy=policy,
                tf_queue_name=tf_queue_name,
                provider=provider)
            self.template.add_aws_sqs_queue_policy(meta=sqs_policy)

    @staticmethod
    def _aws_sqs_queue_policy(tf_queue_name, tf_policy_name, policy,
                              provider=None):
        policy_meta = {
            'queue_url': build_sqs_queue_id_ref(queue_name=tf_queue_name),
            'policy': policy
        }

        if provider:
            policy_meta['provider'] = provider

        resource = {
            tf_policy_name: policy_meta
        }
        return resource

    def _sqs_queue(self, tf_queue_name, fifo_queue, queue_name,
                   delay_seconds,
                   receive_wait_time_seconds, max_message_size,
                   message_retention_seconds=None,
                   content_based_deduplication=None, kms_master_key_id=None,
                   kms_data_key_reuse_period_seconds=None,
                   visibility_timeout_seconds=None,
                   policy=None, redrive_policy=None,
                   provider=None):
        sqs_template = {}

        if fifo_queue:
            sqs_template.update({'fifo_queue': fifo_queue})

        if delay_seconds:
            sqs_template.update({'delay_seconds': delay_seconds})

        if max_message_size:
            sqs_template.update({'max_message_size': max_message_size})

        if queue_name:
            sqs_template.update({'name': queue_name})

        if receive_wait_time_seconds:
            sqs_template.update(
                {'receive_wait_time_seconds': receive_wait_time_seconds})

        if redrive_policy:
            redrive_policy = {
                'deadLetterTargetArn': redrive_policy['deadLetterTargetArn'],
                'maxReceiveCount': redrive_policy['maxReceiveCount']
            }
            sqs_template.update({'redrive_policy': json.dumps(redrive_policy)})

        if content_based_deduplication:
            sqs_template.update(
                {"content_based_deduplication": content_based_deduplication})

        if kms_master_key_id:
            sqs_template.update({'kms_master_key_id': kms_master_key_id})

        if kms_data_key_reuse_period_seconds:
            sqs_template.update({
                'kms_data_key_reuse_period_seconds': kms_data_key_reuse_period_seconds})

        if visibility_timeout_seconds:
            sqs_template.update(
                {'visibility_timeout_seconds': visibility_timeout_seconds})

        if policy:
            sqs_template.update({"policy": policy})

        if message_retention_seconds:
            sqs_template.update(
                {'message_retention_seconds': message_retention_seconds})

        if provider:
            sqs_template['provider'] = provider

        resource = {
            tf_queue_name: sqs_template
        }
        return resource

    @staticmethod
    def _build_redrive_policy(dead_letter_target_arn, max_receive_count):
        resource = {
            "deadLetterTargetArn": dead_letter_target_arn,
            "maxReceiveCount": max_receive_count
        }
        return resource

    @staticmethod
    def build_resource_name(is_fifo, name):
        resource_name = name
        if is_fifo and not name.endswith(FIFO_SUFFIX):
            resource_name += FIFO_SUFFIX
        return resource_name



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import datetime
import os
import sys

import yaml
from botocore.exceptions import ClientError
from boto3.session import Session

from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.connection.sts_connection import STSConnection
from syndicate.core.conf.processor import (
    PROJECT_PATH_CFG, CONFIG_FILE_NAME, ALIASES_FILE_NAME,
    ACCOUNT_ID_CFG, REGION_CFG, DEPLOY_TARGET_BUCKET_CFG,
    AWS_ACCESS_KEY_ID_CFG, AWS_SECRET_ACCESS_KEY_CFG, RESOURCES_PREFIX_CFG,
    RESOURCES_SUFFIX_CFG, TAGS_CFG, IAM_PERMISSIONS_BOUNDARY_CFG,
    TEMP_AWS_ACCESS_KEY_ID_CFG, TEMP_AWS_SECRET_ACCESS_KEY_CFG,
    TEMP_AWS_SESSION_TOKEN_CFG)
from syndicate.core.conf.validator import (LAMBDAS_ALIASES_NAME_CFG,
                                           USE_TEMP_CREDS_CFG,
                                           SERIAL_NUMBER_CFG,
                                           ACCESS_ROLE_CFG,
                                           EXTENDED_PREFIX_MODE_CFG,
                                           LOGS_EXPIRATION, EXPIRATION_CFG,
                                           LOCK_LIFETIME_MINUTES_CFG)
from syndicate.core.constants import DEFAULT_LOGS_EXPIRATION
from syndicate.core.generators import _mkdir

_LOG = get_logger('config_generator')
_USER_LOG = get_user_logger()


def generate_configuration_files(name, config_path, region,
                                 access_key, secret_key,
                                 bundle_bucket_name, prefix, suffix,
                                 extended_prefix, session_token=None,
                                 project_path=None, use_temp_creds=None,
                                 access_role=None, serial_number=None,
                                 tags=None, iam_permissions_boundary=None,
                                 lock_lifetime_minutes=None):
    if not access_key and not secret_key:
        _USER_LOG.warn("Access_key and secret_key weren't passed. "
                       "Attempting to load them")
        credentials = Session().get_credentials()
        if not credentials:
            raise AssertionError("No credentials could be found")

    try:
        sts = STSConnection(region=region,
                            aws_access_key_id=access_key,
                            aws_secret_access_key=secret_key,
                            aws_session_token=session_token)
        caller_identity = sts.get_caller_identity()
        account_id = str(caller_identity['Account'])
    except ClientError:
        _USER_LOG.error('Invalid credentials provided, please '
                        'specify the correct one')
        _LOG.exception('Error while account_id obtaining')
        sys.exit(1)

    if not config_path:
        _LOG.warn(f'The config_path property is not specified. '
                  f'The working directory is used')
        config_path = os.getcwd()

    if not os.path.exists(config_path):
        try:
            os.mkdir(config_path)
        except OSError:
            _LOG.exception(
                f'Error has occurred while creating '
                f'directory for configs {config_path}')
            sys.exit(1)

    config_folder_path = os.path.join(config_path, f'.syndicate-config-{name}')
    _mkdir(path=config_folder_path)

    if not project_path:
        _USER_LOG.warn(f'The "{PROJECT_PATH_CFG}" property is not specified. '
                       f'The working directory will be used as a project path.'
                       f' To change the path, edit the {CONFIG_FILE_NAME} '
                       f'by path {config_folder_path}')
        project_path = os.getcwd()
    else:
        if not os.path.exists(project_path):
            raise AssertionError(
                f'Provided project path {project_path} does not exists')
        project_path = os.path.abspath(project_path)

    if use_temp_creds and access_role:
        raise AssertionError(f'Access role mustn\'t be specified if '
                             f'\'use_temp_creds\' parameter is equal to True')

    config_content = {
        ACCOUNT_ID_CFG: account_id,
        REGION_CFG: region,
        DEPLOY_TARGET_BUCKET_CFG: bundle_bucket_name,
        PROJECT_PATH_CFG: project_path,
        RESOURCES_PREFIX_CFG: prefix,
        RESOURCES_SUFFIX_CFG: suffix,
        EXTENDED_PREFIX_MODE_CFG: extended_prefix,
        USE_TEMP_CREDS_CFG: use_temp_creds,
        ACCESS_ROLE_CFG: access_role,
        SERIAL_NUMBER_CFG: serial_number,
        TAGS_CFG: tags,
        IAM_PERMISSIONS_BOUNDARY_CFG: iam_permissions_boundary,
        LOCK_LIFETIME_MINUTES_CFG: lock_lifetime_minutes
    }
    if session_token:
        config_content.update({
            TEMP_AWS_ACCESS_KEY_ID_CFG: access_key,
            TEMP_AWS_SECRET_ACCESS_KEY_CFG: secret_key,
            TEMP_AWS_SESSION_TOKEN_CFG: session_token,
            EXPIRATION_CFG: (
                    datetime.datetime.now(datetime.timezone.utc) +
                    datetime.timedelta(minutes=50)
            )
        })
    else:
        config_content.update({
            AWS_ACCESS_KEY_ID_CFG: access_key,
            AWS_SECRET_ACCESS_KEY_CFG: secret_key
        })
    config_content = {key: value for key, value in config_content.items()
                      if value is not None}

    config_file_path = os.path.join(config_folder_path, CONFIG_FILE_NAME)
    with open(config_file_path, 'w') as config_file:
        yaml.dump(config_content, config_file)

    aliases_content = {
        ACCOUNT_ID_CFG: account_id,
        REGION_CFG: region,
        LAMBDAS_ALIASES_NAME_CFG: 'dev',
        LOGS_EXPIRATION: DEFAULT_LOGS_EXPIRATION,
    }
    aliases_file_path = os.path.join(config_folder_path, ALIASES_FILE_NAME)
    with open(aliases_file_path, 'w') as aliases_file:
        yaml.dump(aliases_content, aliases_file)

    _USER_LOG.info(
        'Syndicate initialization has been completed. \n'
        f'Set SDCT_CONF:{os.linesep}'
        f'Unix: export SDCT_CONF="{config_folder_path}"{os.linesep}'
        f'Windows: setx SDCT_CONF {config_folder_path}')


def generate_build_project_mapping(mapping_item, build_type):
    type_mapping = ''
    for item in mapping_item:
        type_mapping += f'{build_type}:{item}'
    return type_mapping



"""
    Copyright 2024 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import sys
from pathlib import Path, PurePath

from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core import ProjectState
from syndicate.core.constants import SWAGGER_UI_SPEC_NAME_TEMPLATE, \
    SWAGGER_UI_CONFIG_FILE_NAME
from syndicate.core.generators import _mkdir, _touch, _write_content_to_file
from syndicate.core.generators.contents import SWAGGER_UI_INDEX_FILE_CONTENT, \
     _generate_swagger_ui_config
from syndicate.core.project_state.project_state import BUILD_MAPPINGS

SWAGGER_UI_RUNTIME = 'swagger_ui'
INDEX_FILE_NAME = 'index.html'

_LOG = get_logger('syndicate.core.generators.swagger_ui')
USER_LOG = get_user_logger()


def generate_swagger_ui(name, spec_path, target_bucket, project_path):
    path_to_project = Path(project_path)
    abs_path_to_spec = path_to_spec = Path(spec_path)
    if not Path.exists(path_to_project):
        USER_LOG.info(f'Project "{project_path}" you '
                      f'have provided does not exist')
        return
    if not ProjectState.check_if_project_state_exists(
            project_path=project_path):
        USER_LOG.info(f'Seems that the path {project_path} is not a project')
        return
    if not Path.is_absolute(abs_path_to_spec):
        abs_path_to_spec = Path.joinpath(path_to_project, path_to_spec)
        USER_LOG.info(f'Path to specification file resolved as '
                      f'\'{abs_path_to_spec}\'')
    if not Path.is_file(abs_path_to_spec):
        raise AssertionError(f'Provided specification file \'{spec_path}\' '
                             f'can\'t be resolved! Please provide the correct '
                             f'path.')

    project_state = ProjectState(project_path=project_path)
    src_path = PurePath(project_path,
                        BUILD_MAPPINGS[SWAGGER_UI_RUNTIME],
                        name).as_posix()
    answer = _mkdir(path=src_path,
                    fault_message=f'Swagger UI with name \'{name}\' already '
                                  f'exists.\nOverride? [y/n]')
    if not answer:
        USER_LOG.info(f'Creation of Swagger UI \'{name}\' cancelled')
        sys.exit()

    _create_config_file(src_path=src_path,
                        resource_name=name,
                        spec_path=PurePath(path_to_spec).as_posix(),
                        target_bucket=target_bucket)

    _create_index_file(src_path=src_path,
                       spec_file_name=SWAGGER_UI_SPEC_NAME_TEMPLATE.format(
                           name=name))

    USER_LOG.info(f'Swagger UI \'{name}\' resource files saved to '
                  f'\'{src_path}\'')

    project_state.add_project_build_mapping(runtime=SWAGGER_UI_RUNTIME)
    project_state.save()
    _LOG.info(f'Swagger UI \'{name}\' added to the project')


def _create_config_file(src_path, resource_name, spec_path, target_bucket):
    config_file = PurePath(src_path, SWAGGER_UI_CONFIG_FILE_NAME)
    _touch(config_file)
    config = _generate_swagger_ui_config(resource_name, spec_path,
                                         target_bucket)
    _write_content_to_file(file=config_file, content=config)


def _create_index_file(src_path, spec_file_name):
    index_file_path = PurePath(src_path, INDEX_FILE_NAME)
    _touch(index_file_path)
    _write_content_to_file(file=index_file_path,
                           content=SWAGGER_UI_INDEX_FILE_CONTENT.replace(
                               'spec_file_name', spec_file_name))



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import base64
import os
from time import sleep

from syndicate.commons.log_helper import get_logger
from syndicate.connection.ec2_connection import InstanceTypes
from syndicate.core import ClientError
from syndicate.core.helper import unpack_kwargs, \
    dict_keys_to_capitalized_camel_case
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj, chunks

_LOG = get_logger('syndicate.core.resources.ec2_resource')


class Ec2Resource(BaseResource):

    def __init__(self, ec2_conn, iam_conn, region, account_id) -> None:
        self.ec2_conn = ec2_conn
        self.iam_conn = iam_conn
        self.region = region
        self.account_id = account_id

    def describe_ec2(self, name, meta, response=None):
        if not response:
            response = self.ec2_conn.describe_instances()
        arn = 'arn:aws:ec2:{0}:{1}:instance/{2}'.format(self.region,
                                                        self.account_id,
                                                        response['InstanceId'])
        describe_response = self.ec2_conn.describe_instances(
            [
                {
                    'Name': 'instance-id',
                    'Values': [response['InstanceId']]
                }
            ]
        )
        response['NetworkInterfaces'] = describe_response
        return {
            arn: build_description_obj(response, name, meta)
        }

    def create_ec2(self, args):
        return self.create_pool(self._create_ec2_from_meta, args)

    @unpack_kwargs
    def _create_ec2_from_meta(self, name, meta):
        from syndicate.core import CONF_PATH
        # checking required parameters
        image_id = meta['image_id']
        image_data = self.ec2_conn.describe_image(image_id=image_id)
        if not image_data:
            raise AssertionError('Image id {0} is invalid'.format(image_id))

        instance_type = meta.get('instance_type')
        if not instance_type:
            raise AssertionError('Instance type must be specified')
        if instance_type not in InstanceTypes.from_botocore():
            raise AssertionError(f'Not available instance type: {instance_type}')

        key_name = meta.get('key_name')
        if not self.ec2_conn.if_key_pair_exists(key_name):
            raise AssertionError('There is no key pair with name: {0}'
                                 .format(key_name))

        availability_zone = meta.get('availability_zone')
        subnet = meta.get('subnet_id')
        if availability_zone:
            subnet_filter = {
                'Name': 'availabilityZone',
                'Values': [availability_zone]
            }
            subnet_list = self.ec2_conn.list_subnets(filters=[subnet_filter])
            if subnet and subnet not in \
                    [subnet_ids['SubnetId'] for subnet_ids in subnet_list]:
                raise AssertionError(
                    'There is no available Subnets with name {0} '
                    'in Availability Zone {1}.'
                        .format(subnet, availability_zone))
            if availability_zone not in self.ec2_conn.get_azs():
                raise AssertionError(
                    'There is no Availability Zone with name: {0}'
                        .format(availability_zone))

        security_groups_names = meta.get('security_group_names')
        if security_groups_names:
            sg_meta = self.ec2_conn.describe_security_groups(
                security_groups_names)
            described_sec_groups_names = [security_group['GroupName']
                                          for security_group in sg_meta]
            for security_group_name in security_groups_names:
                if security_group_name not in described_sec_groups_names:
                    raise AssertionError('Security group {0} does not exist'
                                         .format(security_group_name))

        # checking optional parameters
        user_data_file_name = meta.get('userdata_file')
        user_data_content = None
        if user_data_file_name:
            user_data_location = os.path.join(CONF_PATH, user_data_file_name)
            if not os.path.isfile(user_data_location):
                _LOG.warn('There is no user data {0} found by path {1}. '
                          .format(user_data_file_name, CONF_PATH))
            else:
                with open(user_data_location, 'r') as userdata_file:
                    user_data_content = userdata_file.read()

        # describing instance profile by iam role name
        iam_role_name = meta.get('iam_role')
        iam_instance_profile_object = None
        if iam_role_name:
            instance_profiles = self.iam_conn.get_instance_profiles_for_role(
                role_name=iam_role_name)
            if instance_profiles:
                iam_profile_meta = instance_profiles[0]
                iam_instance_profile_arn = iam_profile_meta['Arn']
                iam_instance_profile_object = {
                    'Arn': iam_instance_profile_arn
                }

        # launching instance
        response = self.ec2_conn.launch_instance(
            name=name,
            image_id=image_id,
            instance_type=instance_type,
            key_name=key_name,
            tags_list=meta.get('tags_list'),
            security_groups_names=meta.get('security_group_names'),
            security_group_ids=meta.get('security_group_ids'),
            user_data=user_data_content,
            iam_instance_profile=iam_instance_profile_object,
            subnet_id=meta.get('subnet_id'),
            availability_zone=availability_zone
        )

        if meta.get('disableApiTermination'):
            disable_api_termination = meta.get('disableApiTermination')
            _LOG.debug('Found disableApiTermination '
                       'property: {0}'.format(disable_api_termination))
            if str(disable_api_termination).lower() == 'true':
                self.ec2_conn.modify_instance_attribute(
                    InstanceId=response['InstanceId'],
                    DisableApiTermination={
                        'Value': True
                    }
                )
        _LOG.info('Created EC2 instance %s. '
                  'Waiting for instance network interfaces configuring.', name)
        sleep(30)  # time for vm to become running
        return self.describe_ec2(name, meta, response)

    def remove_ec2_instances(self, args):
        for param_chunk in chunks(args, 1000):
            self.remove_instance_list(param_chunk)

    def remove_instance_list(self, instance_list):
        instance_ids = [x['config']['description']['InstanceId'] for x in
                        instance_list]
        existing_instances_list = []
        for instance_id in instance_ids:
            try:
                self.ec2_conn.modify_instance_attribute(
                    InstanceId=instance_id,
                    DisableApiTermination={'Value': False}
                )
                existing_instances_list.append(instance_id)
            except ClientError as e:
                if 'InvalidInstanceID.NotFound' in str(e):
                    _LOG.warn('Instance %s does not exist', instance_id)
                else:
                    raise e

        if existing_instances_list:
            self.ec2_conn.terminate_instances(
                instance_ids=existing_instances_list)
        _LOG.info('EC2 instances %s were removed.',
                  str(existing_instances_list))

    def describe_launch_template(self, name, meta, response=None):
        if not response:
            response = self.ec2_conn.describe_launch_templates(lt_name=name)
        else:
            response = [response['LaunchTemplate']]
        lt_id = response[0]['LaunchTemplateId']
        return {
            lt_id: build_description_obj(response, name, meta)
        }

    def create_launch_template(self, args):
        return self.create_pool(self._create_launch_template_from_meta, args)

    @unpack_kwargs
    def _create_launch_template_from_meta(self, name, meta):
        lt_data = self._prepare_launch_template_data(meta)
        response = self.ec2_conn.create_launch_template(
            name=name,
            lt_data=dict_keys_to_capitalized_camel_case(lt_data),
            version_description=meta.get('version_description'),
            tag_specifications=dict_keys_to_capitalized_camel_case(
                meta['tag_specifications']) if
            meta.get('tag_specifications') else None
        )
        return self.describe_launch_template(name, meta, response)

    def remove_launch_templates(self, args):
        return self.create_pool(self._remove_launch_template, args)

    @unpack_kwargs
    def _remove_launch_template(self, arn, config):
        try:
            self.ec2_conn.delete_launch_template(lt_id=arn)
            _LOG.info(f"Launch template with ID '{arn}' removed "
                      f"successfully")
        except ClientError as e:
            if 'InvalidLaunchTemplateId.NotFound' in str(e):
                _LOG.warn(f"Launch template with ID '{arn}' not found")
            else:
                raise e

    def update_launch_template(self, args):
        return self.create_pool(self._update_launch_template_from_meta, args)

    @unpack_kwargs
    def _update_launch_template_from_meta(self, name, meta, context):
        lt_data = self._prepare_launch_template_data(meta)
        lt_description = self.ec2_conn.describe_launch_templates(lt_name=name)
        if not lt_description:
            raise AssertionError(
                f"Launch template with name '{name}' not found")
        lt_latest_version = lt_description[0]['LatestVersionNumber']
        self.ec2_conn.create_launch_template_version(
            lt_name=name,
            source_version=str(lt_latest_version),
            lt_data=dict_keys_to_capitalized_camel_case(lt_data),
            version_description=meta.get('version_description')
        )
        response = self.ec2_conn.modify_launch_template(
            lt_name=name,
            default_version=str(lt_latest_version + 1))
        return self.describe_launch_template(name, meta, response)

    def _prepare_launch_template_data(self, meta):
        from syndicate.core import CONFIG
        lt_data = meta['launch_template_data']
        lt_imds = lt_data.pop('imds_support', None)
        image_id = lt_data.get('image_id')
        if image_id:
            image_data = self.ec2_conn.describe_image(image_id=image_id)
            if not image_data:
                raise AssertionError(f'Image id {image_id} is invalid')

        if lt_imds:
            metadata_options = {'http_tokens': 'required'} \
                if lt_imds == 'v2.0' else {'http_tokens': 'optional'}
            lt_data['metadata_options'] = metadata_options

        key_name = lt_data.get('key_name')
        if key_name and not self.ec2_conn.if_key_pair_exists(key_name):
            raise AssertionError(f'There is no key pair with name: {key_name}')

        security_groups_names = lt_data.get('security_groups')
        if security_groups_names:
            sg_meta = self.ec2_conn.describe_security_groups(
                security_groups_names)
            described_sec_groups_names = [security_group['GroupName']
                                          for security_group in sg_meta]
            for security_group_name in security_groups_names:
                if security_group_name not in described_sec_groups_names:
                    raise AssertionError(
                        f'Security group {security_group_name} does not exist')

        security_group_ids = lt_data.get('security_group_ids')
        if security_group_ids:
            sg_meta = self.ec2_conn.describe_security_groups(
                sg_id=security_group_ids)
            described_sec_group_ids = [security_group['GroupId']
                                       for security_group in sg_meta]
            for security_group_id in security_group_ids:
                if security_group_id not in described_sec_group_ids:
                    raise AssertionError(f'Security group with ID '
                                         f'{security_group_id} does not exist')

        iam_role_name = lt_data.pop('iam_role', None)
        if iam_role_name:
            instance_profiles = self.iam_conn.get_instance_profiles_for_role(
                role_name=iam_role_name)
            if instance_profiles:
                iam_profile_meta = instance_profiles[0]
                lt_data['iam_instance_profile'] = {
                    'arn': iam_profile_meta['Arn'],
                    'name': iam_profile_meta['InstanceProfileName']
                }

        user_data_file_path = lt_data.pop('userdata_file', None)
        if user_data_file_path:
            user_data_content = None
            if not os.path.isabs(user_data_file_path):
                user_data_file_path = os.path.join(CONFIG.project_path,
                                                   user_data_file_path)
            if not os.path.isfile(user_data_file_path):
                _LOG.warn(f'There is no user data found by path '
                          f'{user_data_file_path}. ')
            else:
                with open(user_data_file_path, 'r') as userdata_file:
                    user_data_content = userdata_file.read()
            if user_data_content:
                user_data_b = \
                    base64.b64encode(user_data_content.encode("ascii"))
                user_data = user_data_b.decode('ascii')
                lt_data['user_data'] = user_data

        return lt_data



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import STATE_ACTIVITY_TYPE


class StepFunctionActivityGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = STATE_ACTIVITY_TYPE



from syndicate.core.constants import DAX_CLUSTER_TYPE
from syndicate.core.generators.deployment_resources.base_generator import BaseDeploymentResourceGenerator


class DaxClusterGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = DAX_CLUSTER_TYPE
    CONFIGURATION = {
        'node_type': None,
        'iam_role_name': None,
        'replication_factor': 3,
        'security_group_ids': list,
        'availability_zones': list,
        'cluster_endpoint_encryption_type': 'TLS',
        'parameter_group_name': None,
        'subnet_group_name': None,
        'subnet_ids': []
    }


"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from boto3 import client

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger('syndicate.connection.sts_connection')


@apply_methods_decorator(retry())
class STSConnection(object):
    """ STS connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('sts', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new STS connection.')

    def get_temp_credentials(self, role_arn, acc_id, duration=None,
                             serial_number=None, token_code=None):
        """ Get temporary credentials by assuming role
        :param role_arn: str
        :param acc_id: str
        :param duration: int
        :param serial_number: str
        :param token_code: int
        """
        duration = 3600 if not duration else duration
        arn = 'arn:aws:iam::{0}:role/{1}'.format(acc_id, role_arn)
        session_name = "session_{0}".format(acc_id)
        params = dict(
            RoleArn=arn,
            RoleSessionName=session_name,
            DurationSeconds=duration
        )
        if serial_number and token_code:
            params['SerialNumber'] = serial_number
            params['TokenCode'] = token_code
        response = self.client.assume_role(**params)
        return response['Credentials']

    def get_session_token(self, duration=3600, serial_number=None,
                          token_code=None):
        """ Generates temporary AWS credentials

        :param duration: int - duration, in seconds, that the credentials
            should remain valid. From 900 (15 min) to 129600 (36h)
        :param serial_number: str - The identification number of the MFA
            device that is associated with the IAM user who is making the call
        :param token_code: int - The value provided by the MFA device,
            if MFA is required.

        """
        kwargs = {
            'DurationSeconds': duration,
            'SerialNumber': serial_number,
            'TokenCode': token_code
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        response = self.client.get_session_token(**kwargs)
        return response['Credentials']

    def get_caller_identity(self):
        """ Returns details about the IAM identity whose credentials are used to call the API.
        """
        return self.client.get_caller_identity()



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import COGNITO_USER_POOL_TYPE


class CognitoUserPoolGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = COGNITO_USER_POOL_TYPE
    CONFIGURATION = {
        "region": None,  # taking his time
        "password_policy": {
            "minimum_length": 8,
            "require_uppercase": True,
            "require_symbols": True,
            "require_lowercase": True,
            "require_numbers": True,
        },
        "auto_verified_attributes": list,
        "sms_configuration": {
            "sns_caller_arn": None,
            "external_id": None
        },
        "username_attributes": list,
        "custom_attributes": list,
        "client": dict
    }

    def __init__(self, **kwargs):
        self.custom_attributes = kwargs.pop('custom_attributes') or []
        super().__init__(**kwargs)

    def _generate_resource_configuration(self) -> dict:
        result = super()._generate_resource_configuration()
        for name, attr_type in self.custom_attributes:
            result['custom_attributes'].append({'name': name,
                                                'type': attr_type})
        return result
